import numpy as np
import scipy
import skimage
import skimage.feature
import skimage.segmentation
from tqdm.auto import tqdm

# ----------------------- Helper functions ---------------------------------


def maskpk(pk, mask):
    """remove points in nx2 array which are located outside mask
    wp@tl20190709
    """
    return pk[[mask[tuple(p)] > 0 for p in np.round(pk).astype(int)], :]


def disk(s):
    """make a disk shaped structural element to be used with
    morphological functions
    wp@tl20190709
    """
    d = np.zeros((s, s))
    c = (s - 1) / 2.0
    x, y = np.meshgrid(range(s), range(s))
    d2 = (x - c) ** 2 + (y - c) ** 2
    d[d2 < s**2 / 4.0] = 1
    return d


def fill_nan(im):
    """Assigns the value of the nearest finite valued pixel to infinite valued pixels
    wp@tl20190910
    """
    im = im.copy()
    b = np.where(~np.isfinite(im))
    if len(b[0]):
        a = np.where(np.isfinite(im))
        v = scipy.interpolate.griddata(a, im[a[0], a[1]], b, "nearest")
        for i, j, w in zip(b[0], b[1], v):
            im[i, j] = w
    return im


def collectr(cf):
    """Makes a 1d corrfun out of a 2d corrfun
    The result of the maximum of cf at each distance from the center
    wp@tl20191102
    """
    x, y = np.meshgrid(range(cf.shape[0]), range(cf.shape[1]))
    c = [(i - 1) / 2 for i in cf.shape]
    r = np.sqrt((x - c[0]) ** 2 + (y - c[1]) ** 2).flatten()
    idx = np.argsort(r)
    r = np.round(r[idx]).astype("int")
    cf = np.round(cf.flatten()[idx]).astype("int")
    rr = np.unique(r)
    cf = [np.max(cf[r == i]) for i in rr]
    return rr, cf


def corrfft(im, jm):
    """
    usage: d, cfunc = corrfft(images)
    input:
       im, jm: images to be correlated
    output:
       d:      offset (x,y) in px
       cfunc:  correlation function
    """

    im = im.astype(float)
    jm = jm.astype(float)

    im -= np.nanmean(im)
    im /= np.nanstd(im)
    jm -= np.nanmean(jm)
    jm /= np.nanstd(jm)

    im[np.isnan(im)] = 0
    jm[np.isnan(jm)] = 0

    n_y = np.shape(im)[0]
    n_x = np.shape(im)[1]

    cfunc = np.real(
        np.fft.fftshift(np.fft.ifft2(np.fft.fft2(im) * np.conj(np.fft.fft2(jm))))
    )
    y, x = np.unravel_index(np.nanargmax(cfunc), cfunc.shape)

    d = [x - np.floor(n_x / 2), y - np.floor(n_y / 2)]

    # peak at x=nX-1 means xoffset=-1
    if d[0] > n_x / 2:
        d[0] -= n_x
    if d[1] > n_y / 2:
        d[1] -= n_y
    return d, cfunc


def otsu_local(im, res=16):
    # Executes otsu on blocks of an image, then interpolates to give a local threshold
    # im: array with the image
    # res: image will be cut in res x res blocks
    # Image size should be an integer multiple of res!
    #
    # wp@tl20191121
    if res == 1:  # global otsu
        d = im.flatten()
        d = d[np.isfinite(d)]
        if len(np.unique(d)) > 1:
            return skimage.filters.threshold_otsu(d)
        else:
            return d.flatten()[0]
    th = np.zeros((res, res))
    s = im.shape[0] // res
    for i in range(res):
        for j in range(res):
            d = im[i * s : (i + 1) * s, j * s : (j + 1) * s].flatten()
            d = d[np.isfinite(d)]
            if len(d) == 0:
                th[i, j] = np.nan
            elif len(np.unique(d)) > 1:
                th[i, j] = skimage.filters.threshold_otsu(d)
            else:
                th[i, j] = d.flatten()[0]
    th = skimage.transform.resize(th, im.shape, anti_aliasing=False, mode="edge")
    return scipy.ndimage.gaussian_filter(th, im.shape[0] / 8, mode="reflect")


def fill_holes(im):
    """Fill holes (value==0) in each label individually in a faster way
    wp@tl20200120
    """
    im = im.copy()
    z = skimage.morphology.dilation(im) * (im == 0)
    holes_lbl = skimage.measure.label(im == 0)
    lbl_edges = np.unique(
        np.hstack(
            (holes_lbl[:, 0], holes_lbl[:, -1], holes_lbl[0, :], holes_lbl[-1, :])
        )
    )
    lbl_h = list(np.unique(holes_lbl))
    for i in lbl_h:
        if i == 0 or i in lbl_edges:
            continue
        lbl_edge = list(np.unique((holes_lbl == i) * z))
        if 0 in lbl_edge:
            lbl_edge.remove(0)
        if len(lbl_edge) == 1:
            im += lbl_edge[0] * (holes_lbl == i)
    return im


def make_mask(im, threshold, thres, smooth=2.5, min_feature_size=5, dilate=5):
    lb = fill_nan(im)
    if smooth:
        la = scipy.ndimage.gaussian_filter(lb, smooth, mode="nearest")
    else:
        la = lb.copy()
    la[~np.isfinite(im)] = np.nan
    if threshold is None or isinstance(threshold, str) and threshold.lower() == "none":
        th = otsu_local(la, thres)
    else:
        th = float(threshold)
    mask = la > th
    if min_feature_size > 0:
        mask = skimage.morphology.binary_opening(mask, disk(min_feature_size))
    if dilate == 0:
        return mask, la
    elif dilate > 0:
        return skimage.morphology.binary_dilation(mask, disk(dilate)), la
    else:
        return skimage.morphology.binary_erosion(mask, disk(-dilate)), la


# ----------------------- Finally the function itself ---------------------------------


def findcells(
    im,
    imnuc=None,
    cellcolormask=None,
    ccdist=None,
    threshold=None,
    thresholdnuc=None,
    thres=1,
    thresnuc=1,
    smooth=2.5,
    smoothnuc=2.5,
    minfeatsize=5,
    minfeatsizenuc=5,
    dilate=5,
    dilatenuc=5,
    removeborders=True,
    **_kwargs,
):
    """segment cells and nuclei from an image (nxm array)
    wp@tl20190710

    im: 2d numpy array with an image

    optional:
    imnuc: 2d numpy array with an image containing the nuclei, for example with DAPI
    cellcolormask: use a cellmask from another frame to assign the same colors
                   to the cells found by findcells
    ccdist:        (approximate) minimum distance between the centers of cells, good values: yeast: 25,
       mamalian: 150

    optional parameters to debug/finetune findcells:
    threshold:     value used to threshold the image, default: use Otsu
    thres:         divide the image in thres x thres blocks and use a different threshold in each of them,
                   the image size should be an integer multiple of thres
    smooth:        smooth the image before using it
    minfeatsize:   remove features smaller than minfeatsize
    dilate:        >0: make every cell dilate px larger, <0: make them smaller
    thresholdnuc, thresnuc, smoothnuc, minfeatsizenuc, dilatenuc: parameters applying to imnuc
    removeborders: remove any cells (and their nuclei) touching any borders (case with nan's unhandled atm),
        default: True
    """
    im = np.asarray(im).astype("float")
    if imnuc is not None:
        imnuc = np.asarray(imnuc).astype("float")

    if ccdist is None:  # try to determine good ccdist if not given
        if imnuc is None:
            cf = corrfft(im.copy(), im.copy())[1]
        else:
            cf = corrfft(imnuc.copy(), imnuc.copy())[1]
        r, cf = collectr(cf)
        ccdist = r[scipy.signal.find_peaks(-np.array(cf))[0][0]]
        ccdist = np.round(np.clip(ccdist, 10, np.sqrt(np.prod(im.shape)) / 5)).astype(
            "int"
        )

    if imnuc is None:  # only one channel
        mask, la = make_mask(im, threshold, thres, smooth, minfeatsize, dilate)
        pk = skimage.feature.peak_local_max(
            fill_nan(la), min_distance=ccdist // 2, exclude_border=False
        )
        pk = maskpk(pk, mask)
        pk = np.array(sorted([q.tolist() for q in pk])[::-1])
        markers = np.zeros(im.shape, int)
        for i in range(pk.shape[0]):
            if cellcolormask is None:
                markers[pk[i, 0], pk[i, 1]] = i + 1
            else:
                markers[pk[i, 0], pk[i, 1]] = cellcolormask[pk[i, 0], pk[i, 1]]
        cells = skimage.segmentation.watershed(-la, markers=markers, mask=mask)
        cells *= mask  # in Python2, watershed sometimes colors outside the mask
        lbl = list(np.unique(cells))
        if 0 in lbl:
            lbl.remove(0)
        cells = fill_holes(cells)
        nuclei = np.zeros(im.shape)
        for i in tqdm(
            lbl, disable=len(lbl) < 25, leave=False
        ):  # threshold each cell to find its nucleus
            cellmask = cells == i
            cm_la = cellmask * la
            cell = cm_la.flatten()
            cell = cell[cell > 0]
            cell = cell[cell > np.percentile(cell, 25)]
            pxval = np.unique(cell)
            if len(pxval) == 0:
                th = 0
            elif len(pxval) == 1:
                th = cell[0]
            else:
                th = skimage.filters.threshold_otsu(cell)
            nucleus = cm_la > th
            nuclei += float(i) * nucleus
    else:  # extra channel with nuclei
        nuclei, _ = findcells(
            imnuc,
            cellcolormask=cellcolormask,
            ccdist=ccdist,
            threshold=thresholdnuc,
            thres=thresnuc,
            smooth=smoothnuc,
            minfeatsize=minfeatsizenuc,
            dilate=dilatenuc,
            removeborders=False,
        )
        mask, la = make_mask(im, threshold, thres, smooth, minfeatsize, dilate)
        cells = skimage.segmentation.watershed(-la, markers=nuclei, mask=mask)
        cells *= mask  # in Python2, watershed sometimes colors outside the mask
        cells = np.dstack((cells, nuclei)).max(
            2
        )  # make sure each nucleus has a cell body
        cells = fill_holes(cells)

    if removeborders:
        for lbl in np.unique(
            np.hstack((cells[:, 0], cells[:, -1], cells[0, :], cells[-1, :]))
        ):
            cells[cells == lbl] = 0
            nuclei[nuclei == lbl] = 0
    return cells, nuclei



from __future__ import annotations

import contextlib
import io
import pickle
import re
import sys
import warnings
from abc import ABCMeta
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from functools import wraps
from glob import glob
from inspect import Parameter, getfile, signature
from pathlib import Path
from shutil import copyfile
from traceback import format_exc, print_exception
from typing import Any, Callable, Hashable, Sequence, TypeVar

import makefun
import numpy as np
import pandas
import py
import regex
from bidict import bidict
from IPython import embed
from ruamel import yaml

from .io import get_params, pickle_dump, yaml_dump, yaml_load

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

Number = int | float | complex


R = TypeVar("R")

__all__ = [
    "add_extra_parameters",
    "capture_stderr",
    "cfmt",
    "cprint",
    "Crop",
    "Data",
    "df_join",
    "ErrorValue",
    "format_list",
    "get_config",
    "get_slice",
    "ipy_debug",
    "SliceKeepSize",
    "Struct",
    "warn",
    "wraps_combine",
]


@contextlib.contextmanager
def capture_stderr():
    with contextlib.redirect_stderr(io.StringIO()):
        try:
            capture = py.io.StdCaptureFD(out=False, err=True, in_=False)  # noqa
            yield capture
        except Exception:
            raise Exception
        finally:
            capture.reset()


def wraps_combine(
    wrapper: Callable[[Any, ...], Any] | type, ignore: Sequence[str] = None
) -> Callable[[Any, ...], R]:
    """decorator to combine arguments and doc strings of wrapped and wrapper functions,
    *args and/or **kwargs in wrapped will be replaced by the arguments from wrapper,
    duplicate arguments will be left in place in wrapped

    example:

    def wrapper(a, b, c, **kwargs):
        return a + b + c

    @wraps_combine(wrapper)
    def wrapped(a, d, e=45, *args, f=5, **kwargs):
        return d * e * wrapper(*args, **kwargs)

    signature: wrapped(a, d, e, b, c, *, f=5, **kwargs)
    """

    if ignore is None:
        ignore = []

    class WrapsException(Exception):
        pass

    def wrap(wrapped: Callable[[Any, ...], R]) -> Callable[[Any, ...], R]:
        if wrapped.__doc__ and wrapper.__doc__:
            doc = f"{wrapped.__doc__}\n\nwrapping {wrapper.__name__}:\n\n{wrapper.__doc__.lstrip(' ')}"
        elif wrapped.__doc__:
            doc = wrapped.__doc__
        elif wrapper.__doc__:
            doc = wrapper.__doc__
        else:
            doc = None

        try:
            try:
                sig_wrapper = signature(
                    wrapper.__init__ if isinstance(wrapper, type) else wrapper
                )
                sig_wrapped = signature(
                    wrapped.__init__ if isinstance(wrapped, type) else wrapped
                )
            except ValueError:
                raise WrapsException
            z = [(p, p.name, p.kind) for p in sig_wrapped.parameters.values()]
            p1, n1, k1 = zip(*z) if len(z) else ((), (), ())
            z = [
                (p, p.name, p.kind)
                for p in sig_wrapper.parameters.values()
                if (
                    p.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)
                    or p.name not in n1
                )
                and p.name not in ignore
            ]
            p0, n0, k0 = zip(*z) if len(z) else ((), (), ())

            idx0a = (
                k0.index(Parameter.VAR_POSITIONAL)
                if Parameter.VAR_POSITIONAL in k0
                else None
            )
            idx0k = (
                k0.index(Parameter.VAR_KEYWORD) if Parameter.VAR_KEYWORD in k0 else None
            )
            idx1a = (
                k1.index(Parameter.VAR_POSITIONAL)
                if Parameter.VAR_POSITIONAL in k1
                else None
            )
            idx1k = (
                k1.index(Parameter.VAR_KEYWORD) if Parameter.VAR_KEYWORD in k1 else None
            )

            if idx1a is not None:
                if idx0a:
                    new_parameters = [
                        Parameter(p.name, p.kind, annotation=p.annotation)
                        for p in p1[:idx1a]
                    ]
                else:
                    new_parameters = list(p1[:idx1a])
                if (
                    len(new_parameters) == 0
                    or new_parameters[-1].default == Parameter.empty
                ):
                    new_parameters.extend(p0[:idx0k])
                else:
                    new_parameters.extend(
                        [
                            Parameter(
                                p.name,
                                p.kind,
                                default="empty"
                                if p.default == Parameter.empty
                                else p.default,
                                annotation=p.annotation,
                            )
                            for p in p0[:idx0k]
                        ]
                    )
                new_parameters.extend(p1[idx1a + 1 : idx1k])
            elif idx1k is not None:
                if idx0a is not None:
                    new_parameters = list(p1[:idx1k])
                else:
                    new_parameters = [
                        Parameter(p.name, p.kind, annotation=p.annotation)
                        for p in p1[:idx1k]
                    ]
                new_parameters.extend(
                    [
                        Parameter(
                            p.name,
                            Parameter.KEYWORD_ONLY,
                            default="empty"
                            if p.default == Parameter.empty
                            else p.default,
                            annotation=p.annotation,
                        )
                        for p in p0[: (idx0k if idx0a is None else idx0a)]
                    ]
                )
            else:
                new_parameters = list(p1)
            if idx0k is not None:
                new_parameters.append(p0[idx0k])

            @makefun.wraps(
                wrapped, new_sig=sig_wrapper.replace(parameters=new_parameters), doc=doc
            )
            def fun(*args: Any, **kwargs: Any) -> R:
                return wrapped(*args, **kwargs)

        except WrapsException:

            @makefun.wraps(wrapped, doc=doc)
            def fun(*args: Any, **kwargs: Any) -> R:
                return wrapped(*args, **kwargs)

        except Exception:  # noqa
            warnings.warn(
                f"Exception annotating function {wrapped.__name__}:\n\n{format_exc()}"
            )

            @makefun.wraps(wrapped, doc=doc)
            def fun(*args: Any, **kwargs: Any) -> R:
                return wrapped(*args, **kwargs)

        return fun

    return wrap  # type: ignore


class Struct(dict):
    """dict where the items are accessible as attributes"""

    key_pattern = regex.compile(r"(^(?=\d)|\W)")

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.update(*args, **kwargs)

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        if key not in self:
            raise AttributeError(key)
        return self[key]

    def __setitem__(self, key, value):
        super().__setitem__(self.transform_key(key), value)

    def __getitem__(self, key):
        return super().__getitem__(self.transform_key(key))

    def __contains__(self, key):
        return super().__contains__(self.transform_key(key))

    def __deepcopy__(self, memodict=None):
        cls = self.__class__
        copy = cls.__new__(cls)
        copy.update(**deepcopy(super(), memodict or {}))
        return copy

    def __dir__(self):
        return self.keys()

    def __missing__(self, key):
        return None

    @classmethod
    def transform_key(cls, key):
        return cls.key_pattern.sub("_", key) if isinstance(key, str) else key

    def copy(self):
        return self.__deepcopy__()

    def update(self, *args, **kwargs):
        for arg in args:
            if hasattr(arg, "keys"):
                for key, value in arg.items():
                    self[key] = value
            else:
                for key, value in arg:
                    self[key] = value
        for key, value in kwargs.items():
            self[key] = value


yaml.RoundTripRepresenter.add_representer(
    Struct, yaml.RoundTripRepresenter.represent_dict
)


@dataclass
class ErrorValue:
    """format a value and its error with equal significance
    example f'value = {ErrorValue(1.23234, 0.34463):.2g}'
    """

    value: Number
    error: Number

    def __format__(self, format_spec: str) -> str:
        notation = regex.findall(r"[efgEFG]", format_spec)
        notation = notation[0] if notation else "f"
        value_str = f"{self.value:{format_spec}}"
        digits = regex.findall(r"\d+", format_spec)
        if digits:
            digits = int(digits[0])
        else:
            frac_part = regex.findall(r"\.(\d+)", value_str)
            digits = len(frac_part[0]) if frac_part else 0
        if notation in "gG":
            int_part = regex.findall(r"^[-+]?(\d+)", value_str)
            if int_part:
                digits -= len(int_part[0])
            frac_part = regex.findall(r"\.(\d+)", value_str)
            if frac_part:
                zeros = regex.findall(r"^0+", frac_part[0])
                if zeros:
                    digits += len(zeros[0])
        exp = regex.findall(r"[eE]([-+]?\d+)$", value_str)
        exp = int(exp[0]) if exp else 0
        error_str = f"{round(self.error * 10**-exp, digits):{f'.{digits}f'}}"
        split = regex.findall(r"([^eE]+)([eE][^eE]+)", value_str)
        if split:
            return f"({split[0][0]}±{error_str}){split[0][1]}"
        else:
            return f"{value_str}±{error_str}"

    def __str__(self) -> str:
        return f"{self}"


def cfmt(string: str) -> str:
    """format a string for color printing, see cprint"""
    pattern = regex.compile(
        r"(?:^|[^\\])(?:\\\\)*(<)((?:(?:\\\\)*\\<|[^<])*?)(:)([^:]*?[^:\\](?:\\\\)*)(>)"
    )
    fmt_split = regex.compile(r"(?:^|\W?)([a-zA-Z]|\d+)?")
    str_sub = regex.compile(r"(?:^|\\)((?:\\\\)*[<>])")

    # noinspection PyShadowingNames
    def format_fmt(fmt: str) -> str:
        f = fmt_split.findall(fmt)[:3]
        color, decoration, background = f + [""] * max(0, (3 - len(f)))

        t = "KRGYBMCWargybmcwk"
        d = {"b": 1, "u": 4, "r": 7}
        text = ""
        if len(color):
            if color.isnumeric() and 0 <= int(color) <= 255:
                text = f"\033[38;5;{color}m{text}"
            elif not color.isnumeric() and color in t:
                text = f"\033[38;5;{t.index(color)}m{text}"
        if len(background):
            if background.isnumeric() and 0 <= int(background) <= 255:
                text = f"\033[48;5;{background}m{text}"
            elif not background.isnumeric() and background in t:
                text = f"\033[48;5;{t.index(background)}m{text}"
        if len(decoration) and decoration.lower() in d:
            text = f"\033[{d[decoration.lower()]}m{text}"
        return text

    while matches := pattern.findall(string, overlapped=True):
        for match in matches:
            fmt = format_fmt(match[3])
            sub_string = match[1].replace("\x1b[0m", f"\x1b[0m{fmt}")
            string = string.replace("".join(match), f"{fmt}{sub_string}\033[0m")
    return str_sub.sub(r"\1", string)


@wraps_combine(print)
def cprint(*args, **kwargs):
    """print colored text
    text between <> is colored, escape using \\ to print <>
    text and color format in <> is separated using : and text color, decoration and background color are separated
    using . or any character not a letter, digit or :
    colors: 'krgybmcw' (darker if capitalized) or terminal color codes (int up to 255)
    decorations: b: bold, u: underlined, r: swap color with background color"""
    print(*(cfmt(arg) for arg in args), **kwargs)


@wraps(warnings.warn)
def warn(
    message: str, category: type = None, stacklevel: int = 1, source: Any = None
) -> None:
    warnings.warn(cfmt(f"<{message}:208>"), category, stacklevel + 1, source)


def format_list(string: str, lst: Sequence, fmt: str = None) -> str:
    """format a list in a grammatically correct way
    example: format_list('in {channel|channels}: {}', (1, 2, 5))
        'in channels: 1, 2 and 5'
    """
    if fmt is None:
        fmt = ""
    string = string.replace("{}", "{0}")
    plurals = re.findall(r"{([^|{}]+)\|([^|{}]+)}", string)
    for i, option in enumerate(plurals, start=1):
        string = string.replace(f"{{{option[0]}|{option[1]}}}", f"{{{i}}}")
    if len(lst) == 1:
        return string.format(f"{lst[0]:{fmt}}", *[option[0] for option in plurals])
    else:
        return string.format(
            ", ".join([f"{i:{fmt}}" for i in lst[:-1]]) + f" and {lst[-1]:{fmt}}",
            *[option[1] for option in plurals],
        )


def ipy_debug():
    """Enter ipython after an exception occurs any time after executing this."""

    def excepthook(etype, value, traceback):
        print_exception(etype, value, traceback)
        embed(colors="neutral")

    sys.excepthook = excepthook


def get_slice(shape, n):
    ndim = len(shape)
    if isinstance(n, type(Ellipsis)):
        n = [None] * ndim
    elif not isinstance(n, (tuple, list)):
        n = [n]
    else:
        n = list(n)
    ell = [i for i, e in enumerate(n) if isinstance(e, type(Ellipsis))]
    if len(ell) > 1:
        raise IndexError("an index can only have a single ellipsis (...)")
    if len(ell):
        if len(n) > ndim:
            n.remove(Ellipsis)  # type: ignore
        else:
            n[ell[0]] = None
            while len(n) < ndim:
                n.insert(ell[0], None)
    while len(n) < ndim:
        n.append(None)

    pad = []
    for i, (e, s) in enumerate(zip(n, shape)):
        if e is None:
            e = slice(None)
        elif isinstance(e, Number):
            e = slice(e, e + 1)
        if isinstance(e, (slice, range)):
            start = int(np.floor(0 if e.start is None else e.start))
            stop = int(np.ceil(s if e.stop is None else e.stop))
            step = round(1 if e.step is None else e.step)
            if step != 1:
                raise NotImplementedError(
                    "step sizes other than 1 are not implemented!"
                )
            pad.append((max(0, -start) // step, max(0, stop - s) // step))
            if start < 0:
                start = 0
            elif start >= s:
                start = s
            if stop >= s:
                stop = s
            elif stop < 0:
                stop = 0
            n[i] = slice(start, stop, step)  # type: ignore
        else:
            a = np.asarray(n[i])
            if not np.all(a[:-1] <= a[1:]):
                raise NotImplementedError("unsorted slicing arrays are not supported")
            n[i] = a[(0 <= a) * (a < s)]  # type: ignore
            pad.append((sum(a < 0), sum(a >= s)))

    return n, pad


@dataclass
class Crop:
    """Special crop object which never takes data from outside the array, and returns the used extent too,
    together with an image showing how much of each pixel is within the extent,
    negative indices are taken literally, they do not refer to the end of the dimension!
    """

    array: np.ndarray

    def __getitem__(self, n):
        n = get_slice(self.array.shape, n)[0]
        return np.vstack([(i.start, i.stop) for i in n]), self.array[tuple(n)]  # type: ignore


@dataclass
class SliceKeepSize:
    """Guarantees the size of the slice by filling with a default value,
    negative indices are taken literally, they do not refer to the end of the dimension!
    """

    array: np.ndarray
    default: Number = 0

    def __getitem__(self, n):
        n, pad = get_slice(self.array.shape, n)
        crop = self.array[tuple(n)]
        default = self.default(crop) if callable(self.default) else self.default
        return np.pad(crop, pad, constant_values=default)  # type: ignore

    def __setitem__(self, n, value):
        n = np.vstack(n)
        idx = np.prod([(0 < i) & (i < s) for i, s in zip(n, self.array.shape)], 0) > 0
        if not isinstance(value, Number):
            value = np.asarray(value)[idx]
        if n.size:
            self.array[tuple(n[:, idx])] = value


class Data(metaclass=ABCMeta):
    params = None
    do_not_pickle = ()
    channels: bidict
    colors: bidict

    def __init__(self) -> None:
        self.stage = set()
        self.runtime = datetime.now().strftime("%Y%m%d_%H%M%S")

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.error = format_exc()
        self.save()

    @classmethod
    def load(cls, file: Path | str) -> Self:
        files = glob(str(file))
        if len(files) == 0:
            raise FileNotFoundError
        file = Path(max(files)).resolve()
        with open(file, "rb") as f:
            new = pickle.load(f)
        new.__class__ = cls
        new.file = file
        return new

    @staticmethod
    def stage_rec(fun: Callable) -> Callable:
        def wrap(self, *args, **kwargs):
            res = fun(self, *args, **kwargs)
            self.stage.add(fun.__name__)
            return res

        return wrap

    def save(self, file: Path | str = None) -> None:
        if file is None and hasattr(self, "folder_out"):
            file = (
                self.folder_out
                / f"{self.__class__.__name__.lower()}_{self.runtime}.pickle"
            )
        if file is not None:
            pickle_dump(self, file)  # type: ignore

    @classmethod
    def load_from_parameter_file(cls, parameter_file: Path | str) -> Self:
        parameter_file = Path(parameter_file)
        params = getParams(
            parameter_file.with_suffix(".yml"), required=({"paths": ("folder_out",)},)
        )
        if Path(params["paths"]["folder_out"]).exists():
            pickles = [
                file
                for file in Path(params["paths"]["folder_out"]).iterdir()
                if file.name.startswith(f"{cls.__name__.lower()}_")
                and file.suffix == ".pickle"
            ]
        else:
            pickles = None
        if not pickles:
            raise FileNotFoundError(
                f"No files matching {Path(params['paths']['folder_out']) / f'{cls.__name__.lower()}_*.pickle'}"
            )
        return cls.load(max(pickles))

    def run(self) -> None:
        self.runtime = datetime.now().strftime("%Y%m%d_%H%M%S")

    def clean(self) -> None:
        if Path(self.params["paths"]["folder_out"]).exists():
            pickles = [
                file
                for file in Path(self.params["paths"]["folder_out"]).iterdir()
                if file.name.startswith(f"{self.__class__.__name__.lower()}_")
                and file.suffix == ".pickle"
            ]
            if pickles:
                pickles.remove(max(pickles))
                for pkl in pickles:
                    pkl.unlink()

    def color(self, color_or_channel: int | str) -> str:
        return (
            color_or_channel
            if isinstance(color_or_channel, str)
            else self.channels[color_or_channel]
        )

    def channel(self, color_or_channel: int | str) -> int:
        return (
            self.colors[color_or_channel]
            if isinstance(color_or_channel, str)
            else color_or_channel
        )

    @classmethod
    def get_template_file(cls) -> Path:
        return Path(getfile(cls).replace(".py", "_parameters_template.yml"))

    @classmethod
    def copy_template_file(cls) -> None:
        source = cls.get_template_file()
        cwd = Path.cwd()
        stem = source.stem
        suffix = source.suffix
        dest = cwd / source.name

        i = 0
        while dest.exists():
            i += 1
            dest = cwd / f"{stem}_{i}{suffix}"

        copyfile(source, dest)

    @classmethod
    def update_parameter_file(cls, parameter_file: str | Path) -> None:
        parameter_file = Path(parameter_file)
        new_parameter_file = (
            parameter_file.parent / f"{parameter_file.stem}_updated.yml"
        )
        if new_parameter_file.exists():
            raise FileExistsError(f"File {new_parameter_file} already exists.")
        else:
            template = cls.get_template_file()
            yaml_dump(
                get_params(
                    template,
                    parameter_file,
                    replace_values=True,
                    template_file_is_parent=True,
                    compare=True,
                    warn=False,
                ),
                new_parameter_file,
                unformat=True,
            )


def df_join(h: pandas.DataFrame) -> pandas.DataFrame:
    """join DataFrames given by the first indices of h on the other indices"""
    groups = h.groupby(level=0)
    n = len(groups)
    df, j = None, None
    for a, (i, g) in enumerate(groups):
        if a == 0:
            df = g.droplevel(0)
        elif a == n - 1:
            return df.join(g.droplevel(0), lsuffix=f"_{j:.0f}", rsuffix=f"_{i:.0f}")
        else:
            df = df.join(g.droplevel(0), lsuffix=f"_{j:.0f}")
        j = i
    return df


def add_extra_parameters(
    parameters: dict[Hashable, Any], extra_parameters: dict[Hashable, Any]
) -> None:
    for key, value in extra_parameters.items():
        if isinstance(value, dict):
            add_extra_parameters(parameters[key], value)
        else:
            parameters[key] = value


get_config = yaml_load
getConfig = get_config
getParams = get_params
objFromDict = Struct



from __future__ import annotations

import logging
import os
import re
import tempfile
import warnings
from contextlib import ExitStack, redirect_stdout
from functools import wraps
from inspect import getfullargspec
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Sequence

from .misc import capture_stderr

with capture_stderr():
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    import tensorflow

    logger = tensorflow.get_logger()
    logger.setLevel(logging.ERROR)

import numpy as np
import pandas
with redirect_stdout(StringIO()):
    from cellpose import models
from csbdeep.utils import normalize
from ndbioimage import Imread
from numpy.typing import ArrayLike, NDArray
from scipy.interpolate import interp1d
from scipy.ndimage import distance_transform_edt
from scipy.spatial.distance import cdist
from skimage.segmentation import watershed
from stardist.models import StarDist2D  # noqa
from tiffwrite import FrameInfo, IJTiffFile, IJTiffParallel
from tqdm.auto import tqdm, trange

from .findcells import findcells
from .pytrackmate import trackmate_peak_import

try:
    import imagej
    import scyjava

    def kill_vm():
        if scyjava.jvm_started():
            scyjava.shutdown_jvm()

except ImportError:
    imagej = None
    scyjava = None

    def kill_vm():
        return


def label_dist(labels: np.ndarray, lbl: int, mask: np.ndarray = None) -> np.ndarray:
    """make an array with distances to the edge of the lbl in labels, negative outside, positive inside"""
    lbl_mask = labels == lbl
    dist = -distance_transform_edt(lbl_mask == 0)
    dist[dist < 0] += 1
    dist += distance_transform_edt(lbl_mask == 1)
    dist[(labels != lbl) * (labels != 0)] = -np.inf
    if mask is not None:
        dist[mask] = -np.inf
    return dist


def interp_label(
    t: int,
    ts: Sequence[int],
    labels: Sequence[np.ndarray],
    lbl: int,
    mask: np.ndarray = None,
) -> np.ndarray:
    """return a label field with lbl at time q interpolated from labels at times ts"""
    return lbl * (
        interp1d(
            ts,
            np.dstack([label_dist(label, lbl, mask) for label in labels]),
            fill_value=np.zeros_like(labels[0]),
            bounds_error=False,
        )(t)
        > 0
    )


class SwapLabels:
    def __init__(self, tracks: pandas.DataFrame, min_frames: int = None) -> None:
        if min_frames:
            tracks = (
                tracks.groupby("label")
                .apply(lambda df: df.assign(n_frames=len(df)))
                .query("n_frames > @min_frames", local_dict=dict(min_frames=min_frames))
                .reset_index(drop=True)
            )
        # ensure that labels are consecutive
        if not tracks.empty:
            d = {u: i for i, u in enumerate(tracks["label"].unique(), 1)}
            tracks["label"] = tracks.apply(lambda s: d[s["label"]], 1)
        self.tracks = tracks

    def __call__(
        self, im: Imread, frame_in: np.ndarray, c: int, z: int, t: int
    ) -> np.ndarray:
        frame_out = np.zeros_like(frame_in)
        for i, j in self.get_mapping(t).items():
            frame_out[frame_in == i] = j
        return frame_out

    def get_mapping(self, t: int) -> dict[int, int]:
        return {
            int(i): int(j)
            for i, j in self.tracks.query("t == @t", local_dict=dict(t=t))[
                ["median_intensity", "label"]
            ].to_numpy()
            if np.isfinite(i)
        }

    def get_labels(self) -> np.ndarray:
        return self.tracks["label"].unique()


def sort_labels(tracks: pandas.DataFrame) -> pandas.DataFrame:
    """make labels consistent across different runs"""
    relabel_dict = {
        int(key): value
        for value, key in enumerate(
            tracks.groupby("label").aggregate("mean").sort_values("area").index, 1
        )
    }
    return tracks.groupby("label").apply(
        lambda x: x.assign(label=relabel_dict[x["label"].mean()])
    )


def get_time_points(t: int, missing: Sequence[int]) -> tuple[int, int]:
    t_a = t - 1
    while t_a in missing:
        t_a -= 1
    t_b = t + 1
    while t_b in missing:
        t_b += 1
    return t_a, t_b


def interpolate_missing(
    tracks: pandas.DataFrame, t_len: int = None
) -> pandas.DataFrame:
    """interpolate the position of the cell in missing frames"""
    missing = []
    for cell in tracks["label"].unique():
        h = tracks.query("label==@cell", local_dict=dict(cell=cell))
        if t_len is None:
            t_missing = list(
                set(range(int(h["t"].min()), int(h["t"].max()))) - set(h["t"])
            )
        else:
            t_missing = list(set(range(t_len)) - set(h["t"]))
        g = pandas.DataFrame(
            np.full((len(t_missing), tracks.shape[1]), np.nan), columns=tracks.columns
        )
        g["t"] = t_missing
        g["t_stamp"] = t_missing
        g["x"] = np.interp(t_missing, h["t"], h["x"])
        g["y"] = np.interp(t_missing, h["t"], h["y"])
        g["label"] = cell
        missing.append(g)
    return pandas.concat(missing, ignore_index=True)


def substitute_missing(
    tracks: pandas.DataFrame, missing: pandas.DataFrame, distance: int = 1
) -> pandas.DataFrame:
    """relabel rows in tracks if they overlap with a row in missing"""
    for _, row in missing.iterrows():
        a = tracks.query(
            "t==@t & (x-@x)**2 + (y-@y)**2 < @distance",
            local_dict=dict(t=row["t"], x=row["x"], y=row["y"], distance=distance),
        ).copy()
        a["label"] = row["label"]
        if len(a) == 1:
            tracks.loc[a.index[0], "label"] = row["label"]
        elif len(a) > 1:
            idx = ((a[["x", "y"]] - row[["x", "y"]].tolist()) ** 2).sum(1).idxmin()
            tracks.loc[idx, "label"] = row["label"]
    return tracks


def filter_kwargs(function: Callable, kwargs: dict[str, Any]) -> dict[str, Any]:
    args = getfullargspec(function)
    return {
        key: value
        for key, value in kwargs.items()
        if key in args.args + args.kwonlyargs
    }


def get_xy(im: ArrayLike) -> tuple[float, float]:
    im = np.asarray(im)
    x, y = np.meshgrid(np.arange(im.shape[1]), np.arange(im.shape[0]))
    s = np.sum(im)
    return np.sum(im * y) / s, np.sum(im * x) / s


def connect_nuclei_with_cells(nuclei: ArrayLike, cells: ArrayLike) -> np.ndarray:
    nuclei = np.asarray(nuclei)
    cells = np.asarray(cells)
    i_nuclei = np.array([i for i in np.unique(nuclei) if i > 0])
    i_cells = np.array([i for i in np.unique(cells) if i > 0])
    if len(i_nuclei) == 0 | len(i_cells) == 0:
        return np.zeros_like(cells)
    j = (nuclei.flatten()) > 0 | (cells.flatten() > 0)
    nuclei_flat = nuclei.flatten()[j]
    cells_flat = cells.flatten()[j]
    jaccard = cdist(
        np.vstack([nuclei_flat == i for i in i_nuclei]).astype(int),
        np.vstack([cells_flat == i for i in i_cells]).astype(int),
        "jaccard",
    )
    idx = np.where(jaccard < 0.95)
    nuclei_lbl, cell_lbl = np.vstack((i_nuclei[idx[0]], i_cells[idx[1]]))

    d = {}  # find groups of overlapping nuclei and cells
    for n in np.unique(nuclei_lbl):
        if not any([n in k for k in d.keys()]):
            visited_nuclei = set()
            visited_cells = set()
            nuclei_set = {n}
            cell_set = set()
            while len(nuclei_set - visited_nuclei) or len(cell_set - visited_cells):
                for i in nuclei_set - visited_nuclei:
                    cell_set |= set(cell_lbl[nuclei_lbl == i])
                visited_nuclei |= nuclei_set

                for i in cell_set - visited_cells:
                    nuclei_set |= set(nuclei_lbl[cell_lbl == i])
                visited_cells |= cell_set
            d[tuple(nuclei_set)] = tuple(cell_set)

    cells_new = np.zeros_like(cells)
    x, y = np.meshgrid(range(cells.shape[1]), range(cells.shape[0]))
    for n, c in d.items():
        if len(n) == 1:
            cells_new[np.isin(cells, c)] = n
        else:
            mask = np.zeros(cells.shape, bool)
            mask[np.isin(cells, c)] = True
            mask[np.isin(nuclei, n)] = True
            centers = [get_xy(nuclei == i) for i in n]
            dist = np.min([(x - i[1]) ** 2 + (y - i[0]) ** 2 for i in centers], 0)
            markers = np.zeros_like(cells)
            for i, n_c in zip(centers, n):
                markers[int(round(i[0])), int(round(i[1]))] = n_c
            cells_new[mask] = watershed(dist, markers=markers, mask=mask)[mask]
    cells_new[nuclei > 0] = nuclei[nuclei > 0]
    cells_new[np.isin(cells_new, np.setdiff1d(cells_new, nuclei))] = 0
    return cells_new


def trackmate_fiji(
    file_in: Path | str,
    file_out: Path | str,
    fiji_path: Path | str = None,
    channel: int = 0,
    **kwargs: dict[str, [str, int, float, bool]],  # type: ignore
) -> None:
    if fiji_path is not None:
        fiji_path = Path(fiji_path)
    if fiji_path is not None and fiji_path.exists():
        ij = imagej.init(str(fiji_path))
    else:
        ij = imagej.init("sc.fiji:fiji")
    settings = dict(
        file_in=str(file_in),
        file_out=str(file_out),
        TARGET_CHANNEL=1 + channel,
        MIN_AREA=20,
        MAX_FRAME_GAP=2,
        ALTERNATIVE_LINKING_COST_FACTOR=1.05,
        LINKING_MAX_DISTANCE=15.0,
        GAP_CLOSING_MAX_DISTANCE=15.0,
        SPLITTING_MAX_DISTANCE=15.0,
        ALLOW_GAP_CLOSING=True,
        ALLOW_TRACK_SPLITTING=False,
        ALLOW_TRACK_MERGING=False,
        MERGING_MAX_DISTANCE=15.0,
        CUTOFF_PERCENTILE=0.9,
    )
    settings.update(
        {key.upper(): value for key, value in kwargs.items() if key in settings}  # type: ignore
    )
    with open(Path(__file__).parent / "trackmate.jy") as f:
        ij.py.run_script("py", f.read(), settings)
    ij.dispose()


def trackmate(
    max_file: Path | str,
    tif_file: Path | str,
    tiff_out: Path | str,
    table_out: Path | str = None,
    min_frames: int = None,
    remove_borders: bool = False,
    nucleoli_kwargs: dict[str, Any] = None,
    **kwargs: dict[str, str],
) -> None:
    """run trackmate to make sure cells have the same label in all frames, relabel even if there's just one frame,
    to make sure that cell numbers are consecutive
    also remove cells/nuclei touching the border of the image if needed (only if not a time-lapse)
    """

    with ExitStack() as stack:
        mask = stack.enter_context(Imread(tif_file, axes="ctyx", dtype=int))
        if mask.shape["t"] > 1:
            pat = re.compile(r"_ch1", re.IGNORECASE)
            xml_file = tif_file.with_suffix(".xml")
            trackmate_fiji(tif_file, xml_file, channel=1, **kwargs)
            try:
                tracks = trackmate_peak_import(str(xml_file), get_tracks=True)
                tracks.columns = [pat.sub("", column) for column in tracks.columns]
                tracks = tracks[
                    ["t", "t_stamp", "x", "y", "label", "median_intensity", "area"]
                ]
                missing = interpolate_missing(tracks)
                tracks = substitute_missing(tracks, missing)
                tracks = sort_labels(tracks)
                missing = interpolate_missing(tracks)
                tracks = pandas.concat((tracks, missing), ignore_index=True)[
                    ["label", "median_intensity", "t"]
                ]
            except FileNotFoundError:
                warnings.warn("trackmate encountered an error, segmentation will be saved, but not tracked!")
                tracks = []
                for t in range(mask.shape["t"]):
                    cells = np.unique(mask[:, t])
                    cells = cells[cells > 0]
                    tracks.append(np.vstack((np.arange(1, len(cells) + 1), cells, t * np.ones(len(cells)))).T)
                tracks = pandas.DataFrame(np.vstack(tracks), columns=["label", "median_intensity", "t"])
                missing = None
        else:
            cells = np.unique(mask)
            cells = cells[cells > 0]
            tracks = pandas.DataFrame(
                np.vstack((np.arange(1, len(cells) + 1), cells)).T,
                columns=["label", "median_intensity"],
            ).assign(t=0)
            missing = None
        if table_out:
            tracks.to_csv(table_out, sep="\t", index=False)

        if nucleoli_kwargs.get("remove") is True:
            im = stack.enter_context(Imread(max_file, axes="ctyx"))[
                nucleoli_kwargs["channel"]
            ]
        else:
            im = None

        # relabel the labels according to the tracks and also add missing labels by interpolation
        mask.frame_decorator = SwapLabels(tracks, min_frames)
        dtype = (
            "uint8" if mask.frame_decorator.tracks["label"].max() <= 255 else "uint16"
        )
        tif = stack.enter_context(
            IJTiffFile(tiff_out, pxsize=mask.pxsize_um, colormap="glasbey", dtype=dtype)
        )
        for t in trange(
            mask.shape["t"], desc="reordering cells with trackmate", leave=False
        ):
            frame = np.asarray(mask[:, t])
            if im is not None:
                nucleoli = find_nucleoli(
                    im[t],
                    frame[-1],
                    **nucleoli_kwargs,
                )
                frame[0, nucleoli] = 0
            for c in range(frame.shape[0]):
                if missing is not None:
                    missing_t = missing.query("t==@t", local_dict=dict(t=t))
                    for cell in missing_t["label"].unique():
                        time_points = get_time_points(
                            t,
                            missing.query("label==@cell", local_dict=dict(cell=cell))[
                                "t"
                            ].tolist(),
                        )
                        a = interp_label(
                            t,
                            time_points,
                            [mask[c, i] for i in time_points],  # type: ignore
                            int(cell),
                            frame[c] > 0,
                        )
                        frame[c] += a
            if remove_borders and mask.shape["t"] == 1:
                for lbl in np.unique(
                    np.hstack(
                        (
                            frame[:, :, 0],
                            frame[:, :, -1],
                            frame[:, 0, :],
                            frame[:, -1, :],
                        )
                    )
                ):
                    frame[frame == lbl] = 0
                lbls = np.unique(frame)
                lbls = lbls[lbls > 0]
                frame_orig = frame.copy()
                for i, j in enumerate(lbls):
                    frame[frame_orig == j] = i

            for c in range(frame.shape[0]):
                # TODO: transforms
                tif.save(frame[c], c, 0, t)


def run_stardist(
    image: Path | str,
    tiff_out: Path | str,
    channel_cell: int,
    *,
    model_type: str = None,
    table_out: Path | str = None,
    tm_kwargs: dict[str, str] = None,
    rn_kwargs: dict[str, str | float] = None,
) -> None:
    if model_type is None:
        model_type = "2D_versatile_fluo"
    with redirect_stdout(StringIO()):
        model = StarDist2D.from_pretrained(model_type)
    tm_kwargs = tm_kwargs or {}
    rn_kwargs = rn_kwargs or {}

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / "tm.tif"

        with Imread(image, axes="ctyx") as im:
            with IJTiffFile(tif_file, pxsize=im.pxsize_um) as tif:
                for t in tqdm(
                    range(im.shape["t"]),
                    total=im.shape["t"],
                    desc="running stardist",
                    disable=im.shape["t"] < 10,
                ):
                    tif.save(
                        model.predict_instances(normalize(im[channel_cell, t]))[0],  # type: ignore
                        0,
                        0,
                        t,
                    )

        rn_kwargs["channel"] = channel_cell
        trackmate(
            image, tif_file, tiff_out, table_out, nucleoli_kwargs=rn_kwargs, **tm_kwargs
        )


class CellPoseTiff(IJTiffParallel):
    def __init__(
        self,
        model: models.CellposeModel,
        cp_kwargs: dict[str, str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.model = model
        self.cp_kwargs = cp_kwargs or {}
        super().__init__(*args, **kwargs)

    def parallel(self, frame: tuple[ArrayLike, ...]) -> tuple[FrameInfo, ...]:
        if len(frame) == 1:
            cells = self.model.eval(
                np.stack(frame, 0),  # type: ignore
                channel_axis=0,
                channels=[[0, 0]],
                **self.cp_kwargs,
            )[0]
            return ((cells, 0, 0, 0),)
        else:
            cells = self.model.eval(
                np.stack(frame, 0),  # type: ignore
                channel_axis=0,
                channels=[[1, 0]],
                **self.cp_kwargs,
            )[0]
            nuclei = self.model.eval(
                np.stack(frame, 0),  # type: ignore
                channel_axis=0,
                channels=[[2, 0]],
                **self.cp_kwargs,
            )[0]
            cells = connect_nuclei_with_cells(nuclei, cells)
            return (cells, 0, 0, 0), (nuclei, 1, 0, 0)


def run_cellpose_cpu(
    image: Path | str,
    tiff_out: Path | str,
    channel_cell: int,
    channel_nuc: int = None,
    *,
    model_type: str = None,
    table_out: Path | str = None,
    cp_kwargs: dict[str, str] = None,
    tm_kwargs: dict[str, str] = None,
    rn_kwargs: dict[str, str | float] = None,
) -> None:
    cp_kwargs = cp_kwargs or {}
    tm_kwargs = tm_kwargs or {}
    rn_kwargs = rn_kwargs or {}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = models.CellposeModel(gpu=False, model_type=model_type)
    cp_kwargs = filter_kwargs(model.eval, cp_kwargs)

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / "tm.tif"
        with Imread(image, axes="ctyx") as im:
            with CellPoseTiff(model, cp_kwargs, tif_file, pxsize=im.pxsize_um) as tif:
                for t in tqdm(
                    range(im.shape["t"]),
                    total=im.shape["t"],
                    desc="running cellpose",
                    disable=im.shape["t"] < 10,
                ):
                    tif.save(  # type: ignore
                        (im[channel_cell, t],)
                        if channel_nuc is None
                        else (im[channel_cell, t], im[channel_nuc, t]),
                        0,
                        0,
                        t,
                    )
        rn_kwargs["channel"] = channel_cell if channel_nuc is None else channel_nuc
        trackmate(
            image, tif_file, tiff_out, table_out, nucleoli_kwargs=rn_kwargs, **tm_kwargs
        )


def run_cellpose_gpu(
    image: Path | str,
    tiff_out: Path | str,
    channel_cell: int,
    channel_nuc: int = None,
    *,
    model_type: str = None,
    table_out: Path | str = None,
    cp_kwargs: dict[str, str] = None,
    tm_kwargs: dict[str, str] = None,
    rn_kwargs: dict[str, str | float] = None,
) -> None:
    cp_kwargs = cp_kwargs or {}
    tm_kwargs = tm_kwargs or {}
    rn_kwargs = rn_kwargs or {}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = models.CellposeModel(gpu=True, model_type=model_type or "cyto3")
    cp_kwargs = filter_kwargs(model.eval, cp_kwargs)
    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / "tm.tif"
        with Imread(image, axes="ctyx") as im:
            with IJTiffFile(tif_file, pxsize=im.pxsize_um) as tif:
                for t in tqdm(
                    range(im.shape["t"]),
                    total=im.shape["t"],
                    desc="running cellpose",
                    disable=im.shape["t"] < 10,
                ):
                    if channel_nuc is None:
                        frame = np.stack((im[channel_cell, t],), 0)  # type: ignore
                        cells = model.eval(
                            frame,  # type: ignore
                            channel_axis=0,
                            channels=[[0, 0]],
                            **cp_kwargs,
                        )[0]
                        tif.save(cells, 0, 0, t)
                    else:
                        frame = np.stack((im[channel_cell, t], im[channel_nuc, t]), 0)  # type: ignore
                        cells = model.eval(
                            np.stack(frame, 0),  # noqa
                            channel_axis=0,
                            channels=[[1, 0]],
                            **cp_kwargs,
                        )[0]
                        nuclei = model.eval(
                            np.stack(frame, 0),  # noqa
                            channel_axis=0,
                            channels=[[2, 0]],
                            **cp_kwargs,
                        )[0]
                        cells = connect_nuclei_with_cells(nuclei, cells)
                        tif.save(cells, 0, 0, t)
                        tif.save(nuclei, 1, 0, t)
        rn_kwargs["channel"] = channel_cell if channel_nuc is None else channel_nuc
        trackmate(
            image, tif_file, tiff_out, table_out, nucleoli_kwargs=rn_kwargs, **tm_kwargs
        )


@wraps(run_cellpose_cpu)
def run_cellpose(*args, **kwargs) -> None:
    if len(tensorflow.config.list_physical_devices("GPU")):
        run_cellpose_gpu(*args, **kwargs)
    else:
        run_cellpose_cpu(*args, **kwargs)


class FindCellsTiff(IJTiffParallel):
    def __init__(
        self, fc_kwargs: dict[str, str] = None, *args: Any, **kwargs: Any
    ) -> None:
        self.fc_kwargs = fc_kwargs or {}
        super().__init__(*args, **kwargs)

    def parallel(self, frame: tuple[ArrayLike]) -> tuple[FrameInfo, FrameInfo]:
        cell, nucleus = findcells(*frame, **self.fc_kwargs)
        return (cell, 0, 0, 0), (nucleus, 1, 0, 0)


def run_findcells(
    image: Path | str,
    tiff_out: Path | str,
    channel_cell: int,
    channel_nuc: int = None,
    *,
    table_out: Path | str = None,
    fc_kwargs: dict[str, str] = None,
    tm_kwargs: dict[str, str] = None,
    rn_kwargs: dict[str, str | float] = None,
) -> None:
    fc_kwargs = fc_kwargs or {}
    tm_kwargs = tm_kwargs or {}
    rn_kwargs = rn_kwargs or {}
    fc_kwargs = filter_kwargs(findcells, fc_kwargs)

    with tempfile.TemporaryDirectory() as tempdir:
        tif_file = Path(tempdir) / "tm.tif"

        with Imread(image, axes="ctyx") as im:
            with FindCellsTiff(fc_kwargs, tif_file, pxsize=im.pxsize_um) as tif:
                for t in tqdm(
                    range(im.shape["t"]),
                    total=im.shape["t"],
                    desc="running findcells",
                    disable=im.shape["t"] < 10,
                ):
                    assert channel_cell is not None, "channel_cell cannot be None"
                    tif.save(  # type: ignore
                        (im[channel_cell, t],)
                        if channel_nuc is None
                        else (im[channel_cell, t], im[channel_nuc, t]),
                        0,
                        0,
                        t,
                    )

        rn_kwargs["channel"] = channel_cell if channel_nuc is None else channel_nuc
        trackmate(
            image, tif_file, tiff_out, table_out, nucleoli_kwargs=rn_kwargs, **tm_kwargs
        )


class PreTrackTiff(IJTiffParallel):
    def __init__(
        self, shape_yx: tuple[int, int], radius: float, *args: Any, **kwargs: Any
    ):
        self.shape_yx = shape_yx
        self.xv, self.yv = np.meshgrid(*[range(i) for i in shape_yx])
        self.radius = radius
        super().__init__(*args, **kwargs)

    def parallel(self, cxy: tuple[np.ndarray]) -> tuple[FrameInfo]:
        frame = np.zeros(self.shape_yx, int)
        cxy = cxy[0]
        if len(cxy):
            dist = np.round(
                np.min([(self.yv - i[1]) ** 2 + (self.xv - i[2]) ** 2 for i in cxy], 0)
            ).astype(int)

            for i in cxy:
                frame[int(i[1]), int(i[2])] = i[0]  # noqa
            frame = watershed(dist, frame, mask=dist < self.radius**2)
        return ((frame, 0, 0, 0),)


def run_pre_track(
    image: Path | str, tiff_out: Path | str, pre_track: pandas.DataFrame, radius: float
) -> None:
    dtype = "uint8" if pre_track["cell"].max() < 255 else "uint16"
    with Imread(image) as im:
        with PreTrackTiff(
            im.shape["yx"],  # type: ignore
            radius,
            tiff_out,
            pxsize=im.pxsize_um,
            colormap="glasbey",
            dtype=dtype,
        ) as tif:
            for t in tqdm(
                range(im.shape["t"]),
                total=im.shape["t"],
                desc="running pre track cell masking",
                disable=im.shape["t"] < 10,
            ):
                tif.save(  # type: ignore
                    (
                        pre_track.query("T == @t", local_dict=dict(t=t))[
                            ["cell", "y", "x"]
                        ].to_numpy(),
                    ),
                    0,
                    0,
                    t,
                )


def find_nucleoli(
    image: ArrayLike[Any], mask: ArrayLike[int], **kwargs
) -> NDArray[bool]:
    def fill_mask(img: ArrayLike[Any], msk: ArrayLike[bool]) -> tuple[tuple[int, int, int, int], NDArray[Any]]:
        w = np.where(msk)
        i0, j0 = np.min(w, 1)  # noqa
        i1, j1 = np.max(w, 1)  # noqa
        img = np.asarray(img[i0 : i1 + 1, j0 : j1 + 1]).copy()
        msk = np.asarray(msk[i0 : i1 + 1, j0 : j1 + 1]).copy()
        dist = img[msk]
        a, b = np.percentile(dist, (25, 90))
        dist = dist[(a <= dist) & (dist <= b)]
        msk[img < a] = False
        img[~msk] = np.random.choice(dist, (~msk).sum())
        return (i0, j0, i1, j1), img

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model = models.CellposeModel(gpu=True)
    labels = [label for label in np.unique(mask) if label > 0]
    if labels:
        i, images = zip(
            *[fill_mask(image, mask == label) for label in np.unique(mask) if label > 0]
        )
        kwargs['max_size_fraction'] = 1.0
        masks = model.eval(list(images), **filter_kwargs(model.eval, kwargs))[0]
        mask = np.zeros_like(mask, dtype=bool)
        for (i0, j0, i1, j1), m in zip(i, masks):
            mask[i0 : i1 + 1, j0 : j1 + 1] |= m > 0
        return mask
    else:
        return np.zeros_like(mask, bool)
